from collections import Counter
import numpy as np
from cdp_ts.utils.dataset import Dataset
import numba
from numba import NumbaWarning
import warnings
warnings.filterwarnings("ignore", category=NumbaWarning)


def try_except(func):
    """ Used to decorate functions with try/except clause """
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            print(f"ERROR: {e}")
            result = None
            raise e
        return result
    return wrapper


@numba.jit()
def similarity_coefficient(s1: str, s2: str) -> float:
    """ Calculates similarity between two strings"""

    if len(s1) != len(s2):
        raise ValueError("The patterns have different length!")

    s1_array = np.array(list(s1))
    s2_array = np.array(list(s2))

    matches = np.sum(s1_array == s2_array)
    similarity = matches / len(s1)

    return similarity


@numba.jit()
def entropy(values: np.array) -> float:
    """ Calculates entropy of given sequence"""

    num_values = len(values)
    hist = Counter(values)

    values_array = np.array(list(hist.values()))
    probabilities = values_array / num_values

    entropy = -np.sum(probabilities * np.log(probabilities))

    return entropy


@numba.jit()
def assess_candidate_position(candidate_position: np.array
                              , train_dataframe: Dataset) -> tuple:

    """ Check the fitness of candidate shapelet"""

    distances = []

    for class_index, values in train_dataframe.iterrows():
        distance = subsequent_distance(values, candidate_position)
        distances.append((class_index, distance))

    distances_array = np.array(distances)
    sorted_distances = distances_array[distances_array[:, 1].argsort()]

    information_gain, split_point, optimal_entropy = calculate_information_gain(sorted_distances)

    return information_gain, split_point


@numba.jit()
def subsequent_distance(time_series_values: np.array
                        , candidate_position: np.array) -> float:
    """ Calculates the distance between two time series """

    min_distance = np.inf

    for i in range(len(time_series_values) - len(candidate_position) + 1):
        current_values = time_series_values[i:i + len(candidate_position)]
        squared_diff_sum = np.sum(np.square(current_values - candidate_position))
        min_distance = min(min_distance, squared_diff_sum)

    return min_distance


@numba.jit()
def calculate_information_gain(distances_array: np.array) -> tuple:
    """ Calculate information gain and optimal split distance from list of tuples in format
        (class_index, distance between the time series and the shapelet)"""

    information_gain = 0.0
    optimal_split_distance = 0.0
    optimal_entropy = -1.0

    #distances_array = np.array(distances)
    indexes = distances_array[:, 0]
    distances_values = distances_array[:, 1]

    I = entropy(indexes)

    prev_distance = distances_values[0]

    for i in range(1, len(distances_array)):
        distance = distances_values[i]
        if distance > prev_distance:
            d = (prev_distance + distance) / 2.0
            prev_distance = distance

            h1_indexes = indexes[:i][distances_values[:i] <= d]
            h2_indexes = indexes[i:][distances_values[i:] > d]

            I1 = entropy(h1_indexes)
            I2 = entropy(h2_indexes)

            f1 = len(h1_indexes) / len(indexes)
            f2 = len(h2_indexes) / len(indexes)

            current_entropy = f1 * I1 + f2 * I2
            current_information_gain = I - current_entropy

            if current_information_gain > information_gain:
                information_gain = current_information_gain
                optimal_split_distance = d
                optimal_entropy = current_entropy

    return information_gain, optimal_split_distance, optimal_entropy


def process_dataset(dataset: Dataset
                    , compression_factor: int
                    , normalize: bool
                    , derivative: bool) -> Dataset:
    """ Process given dataframe with compression, normalization or extract derivative """

    if dataset is None or dataset.empty:
        return Dataset()

    dataset = dataset.copy()

    # Apply compression on time series
    if compression_factor > 1:
        dataset.apply_compression(compression_factor)

    # Extract derivative from time series
    if derivative:
        dataset.apply_derivative()

    # Apply normalization on time series
    if normalize:
        dataset.apply_normalization()

    return dataset










import logging

level = logging.INFO

# Create the logger
logger = logging.getLogger(__name__)
logger.setLevel(level)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(level)

# Create a log message format
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console_handler.setFormatter(formatter)

# Add the handlers to the logger
logger.addHandler(console_handler)



""" Concatenated Decision Paths (CDP) main class"""

import os.path
import pickle
import csv
from collections import defaultdict
from cdp_ts.utils.utils import try_except, similarity_coefficient
from cdp_ts.utils.logger import logger
from cdp_ts.utils.dataset import Dataset
from cdp_ts.core.shapelet_classifier import ShapeletClassifier

# Filename of trained model - contains sequence of decision trees
MODEL_FILENAME = 'cdp_model.pickle'
# Filename of csv file that contains predicted class indexes
PATTERNS_FILE_NAME = 'patterns.csv'


class CDP:
    """ Concatenated Decision Paths (CDP) method implementation"""

    def __init__(self
                 , model_folder: str
                 , num_classes_per_tree: int
                 , num_trees: int
                 ):

        self.model_folder = model_folder
        self.num_classes_per_tree = num_classes_per_tree
        self.num_trees = num_trees
        self.patterns = []
        self.classification_trees = {}

    def _load_patterns(self, csv_file_path: str):
        """ Load class indexes along with their representative patterns (LLR..)"""

        # Initialize an empty list to store the loaded data
        self.patterns = []

        # Read the data from the CSV file
        with open(csv_file_path, mode='r', encoding='utf-8') as file:
            reader = csv.DictReader(file)

            # Assuming the column headers are 'class_index' and 'class_pattern'
            for row in reader:
                class_index = int(row['class_index'])
                class_pattern = row['class_pattern']
                self.patterns.append((class_index, class_pattern))

    def _save_patterns(self, csv_file_path: str):
        """ Save class indexes along with their representative patterns (LLR..) """

        # Load patterns from file
        headers = ['class_index', 'class_pattern']

        # Write the data to the CSV file
        with open(csv_file_path, mode='w', encoding='utf-8', newline='') as file:
            writer = csv.writer(file)

            # Write the header row
            writer.writerow(headers)

            # Write the data rows
            for row in self.patterns:
                writer.writerow(row)

    def _save_classification_trees(self, model_folder_path: str):
        """ Save model with classification tree sequence """

        with open(os.path.join(model_folder_path, MODEL_FILENAME), 'wb') as file:
            pickle.dump(self.classification_trees, file)

    def _load_classification_trees(self, model_folder_path: str) -> dict:
        """ Read model file with decision trees sequence """

        try:
            with open(os.path.join(model_folder_path, MODEL_FILENAME), 'rb') as file:
                self.classification_trees = pickle.load(file)
        except Exception as e:
            self.classification_trees = dict()

    @try_except
    def load_model(self):
        """ Load model- classification trees, along with decision patterns """
        # Load model
        self._load_classification_trees(self.model_folder)

        # Load saved patterns
        self._load_patterns(os.path.join(self.model_folder, PATTERNS_FILE_NAME))


    @try_except
    def fit(self, train_dataset: Dataset):

        """ Fills the dictionary with classification trees. If model exists, tries to reuse
        trees if compatible with input arguments requirements"""

        logger.info("Training...")

        shapelet_classifier = ShapeletClassifier(dataset=train_dataset
                                                 , classifiers_folder=self.model_folder
                                                 , num_classes_per_tree=self.num_classes_per_tree
                                                 , num_trees=self.num_trees)

        # Load existing classifiers
        self._load_classification_trees(self.model_folder)

        # Add new classifiers, if required
        shapelet_classifier.create_and_train_classifiers(self.classification_trees)

        # Save all classifiers
        self._save_classification_trees(self.model_folder)

        # Create patterns in format:
        #  [(0, 'LLRLL...LLRLLL')
        # ....
        # , (1, 'RRRRL...RRRRLL')]
        for class_index, time_series in train_dataset.iterrows():
            decision_pattern = ''.join([classification_tree.build_classification_pattern(time_series)
                                       for classification_tree in self.classification_trees.values()]
                                       )
            self.patterns.append((class_index, decision_pattern))

        # Save patterns
        self._save_patterns(os.path.join(self.model_folder, PATTERNS_FILE_NAME))

    @try_except
    def predict(self
                , dataset: Dataset  # pd.DataFrame
                ) -> list:

        """ Predict indexes of given time series dataset"""

        logger.info("Predicting...")

        predicted_class_indexes = []

        # Classify by comparing decision patterns
        for _, time_series in dataset.iterrows():

            # Find the pattern for given time series
            pattern = ''.join([classification_tree.build_classification_pattern(time_series)
                               for classification_tree in self.classification_trees.values()])

            # Find similarity between found pattern and saved during training
            similarities = [(i, similarity_coefficient(pattern, s)) for i, s in self.patterns]

            # Sort tuples by closest distance and take 10 closest
            biggest_similarities = sorted(similarities, key=lambda x: x[1], reverse=True)[:10]

            # Find most popular index among the closest distances
            freq_dict = defaultdict(int)
            for tup in biggest_similarities:
                freq_dict[tup[0]] += 1

            # Finding index with maximum frequency
            predicted_class_indexes.append(max(freq_dict, key=freq_dict.get))

        return predicted_class_indexes



import numpy as np
from math import isclose
from cdp_ts.utils.logger import logger


class Shapelet:
    def __init__(self, values: np.array
                     , optimal_split_distance: float
                     , best_information_gain: float):
        self.values = values
        self.optimal_split_distance = optimal_split_distance
        self.best_information_gain = best_information_gain
        self._left_class_index = -1
        self._right_class_index = -1

    @property
    def left_class_index(self):
        return self._left_class_index

    @left_class_index.setter
    def left_class_index(self, value):
        if value is None:  # and type(value) is not int:
            raise Exception('Left class index is incorrect!')
        self._left_class_index = value

    @property
    def right_class_index(self):
        return self._right_class_index

    @right_class_index.setter
    def right_class_index(self, value):
        if value is None and type(value) is not int:
            raise Exception('Right class index is incorrect!')
        self._right_class_index = value

    def __eq__(self, other):
        if not isinstance(other, Shapelet):
            raise Exception("Not correct Shapelet instance!")

        return \
            np.array_equiv(self.values, other.values)\
            and isclose(self.optimal_split_distance, other.optimal_split_distance)\
            and isclose(self.best_information_gain, other.best_information_gain)\
            and self.left_class_index == other.left_class_index\
            and self.right_class_index == other.right_class_index

    def compare(self, other):

        if not isinstance(other, Shapelet):
            raise Exception("Not correct Shapelet instance!")

        if self == other:
            return 0
        elif self.left_class_index == other.left_class_index\
             or self.left_class_index == other.right_class_index:
            return -1
        elif self.right_class_index == other.left_class_index\
             or self.right_class_index == other.right_class_index:
            return 1
        else:
            return -2

    def __repr__(self):

        logger.info(f'values: {self.values}')
        logger.info(f'optimal_split_distance: {self.optimal_split_distance}')
        logger.info(f'best_information_gain: {self.best_information_gain}')
        logger.info(f'left_class_index: {self.left_class_index}')
        logger.info(f'right_class_index: {self.right_class_index}')



import os
from collections import Counter
from itertools import chain, combinations, permutations
import numpy as np
from cdp_ts.core.shapelet import Shapelet
from cdp_ts.core.pso import ShapeletsPso
from cdp_ts.utils.utils import subsequent_distance
from cdp_ts.utils.btree import BTree
from cdp_ts.utils.logger import logger
from cdp_ts.utils.dataset import Dataset
import concurrent.futures


class ShapeletClassifier:

    def __init__(self
                 , dataset: Dataset
                 , classifiers_folder: str
                 , num_classes_per_tree: int
                 , num_trees: int):

        self.classifiers_folder = classifiers_folder
        self.num_classes_per_tree = num_classes_per_tree
        self.num_trees = num_trees
        self.balanced_dataset = self._get_balanced_dataset(dataset)

    @staticmethod
    def _get_balanced_dataset(dataset: Dataset) -> Dataset:
        """ Create dataset which contains 10 samples from each class with replacement
        """

        # Create a dataset containing 10 samples from each class with replacement.
        balanced_class_indexes = []
        balanced_values = []

        unque_class_indexes = np.unique(dataset.class_indexes)
        num_samlpes = min(10, len(unque_class_indexes))
        for class_index in unque_class_indexes:
            # Get the indices of the current class
            class_indices = np.where(dataset.class_indexes == class_index)[0]

            # Sample 'num_samples_per_class' indices with replacement
            sampled_indices = np.random.choice(class_indices, num_samlpes, replace=True)

            # Append the sampled class indexes and values to the balanced dataset
            balanced_class_indexes.extend([class_index] * num_samlpes)
            balanced_values.extend(dataset.values[sampled_indices])

        # Create the new balanced dataset
        return Dataset(None
                       , None
                       , class_indexes=np.array(balanced_class_indexes)
                       , values=np.array(balanced_values))

    @staticmethod
    def _build_tree(permutation: tuple):
        """ Build tree from given permutation """

        # Initialize the decision tree
        tree = BTree(permutation[0])

        # Build the tree
        for p in permutation:
            tree.add(tree.root, p)

        return tree

    @staticmethod
    def _count_left_right(shapelet: Shapelet, class_dataset: Dataset):

        # Take number of time series which distance to the shapelet is less than optimal split distance
        count_left = 0
        for values in class_dataset.values:
            if subsequent_distance(values, shapelet.values) < shapelet.optimal_split_distance:
                count_left += 1

        return count_left, class_dataset.values.shape[0] - count_left

    def _split_classes(self
                       , shapelet: Shapelet
                       , class_index_a: int
                       , class_index_b: int
                       , dataset: Dataset
                       ):

        """ Assign left and right class index to given shapelet"""

        class_dataset_a = dataset.filter_by_class(class_index_a)
        count_left_a, count_right_a = self._count_left_right(shapelet, class_dataset_a)

        class_dataset_b = dataset.filter_by_class(class_index_b)
        count_left_b, count_right_b = self._count_left_right(shapelet, class_dataset_b)

        # Assign class index, for which majority of samples have
        shapelet.left_class_index = class_index_a if count_left_a > count_left_b else class_index_b
        shapelet.right_class_index = class_index_a if count_right_a > count_right_b else class_index_b

    def _find_shapelet(self, class_index_a: int, class_index_b: int) -> Shapelet:
        """ Find shapelet and its parameters such as optimal split distance, left and right class index"""

        class_indices = np.concatenate(
            (np.where(self.balanced_dataset.class_indexes == class_index_a)[0]
             , np.where(self.balanced_dataset.class_indexes == class_index_b)[0])
            )
        train_dataset = Dataset(class_indexes=self.balanced_dataset.class_indexes[class_indices]
                                , values=self.balanced_dataset.values[class_indices])

        # Start PSO algorithm to find shapelet that separates two classes
        min_length = 3
        max_length = min(len(x) for x in train_dataset.values)
        step = (max_length - min_length)//20
        step = step if step > 0 else 1
        num_classes = len(np.unique(self.balanced_dataset.class_indexes))
        step = step if num_classes >= 4 else 1
        min_train_value = np.min(train_dataset.values[0])
        max_train_value = np.max(train_dataset.values[0])
        shapelet_pso = ShapeletsPso(min_length=min_length
                                    , max_length=max_length
                                    , step=step
                                    , min_position=min_train_value
                                    , max_position=max_train_value
                                    , min_velocity=min_train_value
                                    , max_velocity=max_train_value
                                    , train_dataframe=train_dataset)
        shapelet_pso.start_pso()

        # Fill shapelet parameters- shapelet values, the best info gain, optimal split distance
        shapelet = Shapelet(values=shapelet_pso.best_particle.position[:shapelet_pso.best_particle.length]
                            , best_information_gain=shapelet_pso.best_particle.best_information_gain
                            , optimal_split_distance=shapelet_pso.best_particle.optimal_split_distance)

        # Fill shapelet parameters-left and right class indexes of the shapelet
        self._split_classes(shapelet, class_index_a, class_index_b, train_dataset)

        return shapelet

    def _get_required_classificators_names(self, group: list):

        required_classificators_names = {}
        for combination in group:

            # Check if classifier with such name already exists
            base_name = os.path.join(
                self.classifiers_folder
                , "Classificatin_tree_" + "_".join([str(x) for x in combination]))
            extension = '.pickle'

            # Make unique filename
            counter = 1
            unique_filename = base_name + extension
            while unique_filename in required_classificators_names:
                unique_filename = f"{base_name}_({counter}){extension}"
                counter += 1

            required_classificators_names[unique_filename] = combination

        return required_classificators_names

    def _test_tree_accuracy(self, tree: BTree, classes_in_combination: tuple):

        """ Finds average accuracy of given classification tree """

        for class_index in classes_in_combination:
            # Average accuracy
            acc = 0.0

            # Extract the time series with given class index
            class_indices = np.where(self.balanced_dataset.class_indexes == class_index)[0]
            sampled_indices = np.random.choice(class_indices, 10) # , replace=True
            train_datasets_values = self.balanced_dataset.values[sampled_indices]

            # Number of correctly classified time series for given index
            num_correctly_classified = 0

            for time_series in train_datasets_values:
                current_node = tree.root

                # Iterate the tree
                while True:
                    if current_node is None:
                        break
                    distance = subsequent_distance(time_series, current_node.shapelet.values)

                    # Left wing
                    if distance <= current_node.shapelet.optimal_split_distance:
                        if current_node.left is not None:
                            current_node = current_node.left
                            continue
                        if current_node.shapelet.left_class_index == class_index:
                            num_correctly_classified += 1
                        break

                    # Right wing
                    if distance > current_node.shapelet.optimal_split_distance:
                        if current_node.right is not None:
                            current_node = current_node.right
                            continue
                        if current_node.shapelet.right_class_index == class_index:
                            num_correctly_classified += 1
                        break

            acc += num_correctly_classified/train_datasets_values.shape[0]

        return acc

    def _find_most_accurate_tree(self, shapelets: list, classes_in_combination: tuple):

        """ Tries variety of combinations to build most accurate tree"""

        best_acc = 0.0
        best_tree: BTree = None

        for combination in list(combinations(shapelets, len(classes_in_combination)-1)):
            for permutation in list(permutations(combination)):
                tree = self._build_tree(permutation)
                acc = self._test_tree_accuracy(tree, classes_in_combination)
                if best_acc < acc:
                    best_acc = acc
                    best_tree = tree
                    best_tree.accuracy = acc

        return best_tree

    def _create_group(self) -> list:

        """ Create combination of all class indexes such that all class indexes are uniformly represented """

        class_indexes = np.unique(self.balanced_dataset.class_indexes)
        num_class_indexes = len(class_indexes)
        num_allowed_indexes = self.num_trees * self.num_classes_per_tree // num_class_indexes

        # Generate all possible combinations of 3 classes from the range 0-15
        combs = list(combinations(class_indexes, self.num_classes_per_tree))

        # Filter the combinations to only keep those where each class appears equally
        valid_combs = []
        num_iterations = 0
        # The max number of iterations is set to 2000 artificially as in some cases
        # the output cannot contain all the numbers uniformly distributed and the cycle never ends.
        while not valid_combs \
            or any(freq[class_] < num_allowed_indexes for class_ in list(chain(*valid_combs)))\
                and num_iterations < 2000:
            num_iterations += 1
            for comb in combs:
                freq = Counter(list(chain(*valid_combs)))
                if not valid_combs:
                    valid_combs.append(comb)
                else:
                    if all(freq[class_] < num_allowed_indexes for class_ in comb)\
                            and any(freq[class_] == min(freq.values()) for class_ in comb):
                        valid_combs.append(comb)
                # The numbers in combinations can  be almost uniformly distributed-
                # aka one or two numbers might differ

        return valid_combs

    def _create_and_train_tree(self, classes_in_combination: tuple) -> BTree:
        """ Find shapelets for every pair of classes in given combination. Build the classification
        tree and serialize the most accurate tree"""
        shapelets = []

        # Split combination of classes to all possible two classes combinations
        two_classes_combinations = list(combinations(classes_in_combination, 2))

        # Collect shapelets that split combinations of two classes
        for c in two_classes_combinations:
            class_index_a, class_index_b = c
            shapelet = self._find_shapelet(class_index_a, class_index_b)

            # Find shapelet that split the two classes
            if shapelet.right_class_index is not None\
               and shapelet.left_class_index is not None\
               and shapelet.right_class_index != shapelet.left_class_index:
                shapelets.append(shapelet)

        # Find tree, that gives the best accuracy for given combination of class indexes
        best_tree = self._find_most_accurate_tree(shapelets, classes_in_combination)

        return best_tree

    def create_and_train_classifiers(self, classification_trees: dict):

        """ Create, train group of classifiers defined by input arguments """

        # Find needed combination of classes, defined by input arguments
        group = self._create_group()
        logger.debug(group)

        # Get newly required classifiers- filename: class combination
        required_trees = self._get_required_classificators_names(group)

        # Define trees to be created
        classifiers_file_names = [x for x in required_trees if x not in classification_trees]

        with concurrent.futures.ProcessPoolExecutor() as executor:
            futures = {executor.submit(self._create_and_train_tree, required_trees[classifier_file_name]):
                       classifier_file_name for classifier_file_name in classifiers_file_names}
            for future in concurrent.futures.as_completed(futures):
                classifier_file_name = futures[future]
                try:
                    result = future.result()
                    classification_trees[classifier_file_name] = result
                except Exception as e:
                    print(f"An error occurred for name {classifier_file_name}: {e}")


















from cdp_ts.utils.utils import subsequent_distance
from cdp_ts.core.shapelet import Shapelet
import numpy as np


class BTree:

    class Node:
        def __init__(self, shapelet: Shapelet):
            self.shapelet: Shapelet = shapelet
            self.left = None
            self.right = None
            self.depth = 0

    def __init__(self, shapelet: Shapelet):
        self.root: BTree.Node = self.Node(shapelet)
        self.accuracy: float = 0.0
        self.num_nodes = 1

    def build_classification_pattern(self, time_series: np.array) -> str:

        """ Build classification pattern that represent given time series"""
        current_node: BTree.Node = self.root
        path_string: str = ''

        # Iterate the tree
        while True:
            if current_node is None:
                break

            distance = subsequent_distance(time_series
                                           , current_node.shapelet.values)

            # Left wing
            if distance <= current_node.shapelet.optimal_split_distance:
                path_string += 'L'
                if current_node.left is not None:
                    current_node = current_node.left
                    continue
                else:
                    break
            # Right wing
            else:
                path_string += 'R'
                if current_node.right is not None:
                    current_node = current_node.right
                    continue
                else:
                    break

        # Pad the answer
        while len(path_string) < self.num_nodes:
            path_string += "0"
        return path_string

    def add(self, node: Node, shapelet: Shapelet) -> bool:
        """ Add new node to the tree"""

        add_result: bool = False

        if node is None or shapelet is None:
            return False

        compare_result = node.shapelet.compare(shapelet)

        if compare_result == 0:
            return True

        elif compare_result == -1:
            if node.left is None:
                node.left = BTree.Node(shapelet)
                node.depth += 1
            self.num_nodes += 1
            return self.add(node.left, shapelet)

        elif compare_result == 1:
            if node.right is None:
                node.right = BTree.Node(shapelet)
                node.depth += 1
            self.num_nodes += 1
            return self.add(node.right, shapelet)

        elif compare_result == -2:
            add_result = self.add(node.left, shapelet)
            if not compare_result:
                self.num_nodes += 1
                add_result = self.add(node.right, shapelet)

        return add_result








"""A python implementation of Concatenated-Decision-Path method for time series classification."""
import cdp_ts
from setuptools import find_packages, setup


NAME = 'cdp_ts'
DESCRIPTION = 'A python implementation of Concatenated-Decision-Path method for time series classification'
with open('README.md') as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'
MAINTAINER = 'Ivan Mitzev'
MAINTAINER_EMAIL = 'cdp_project@outlook.com'
URL = 'https://github.com/naviveztim/CDP_python'
LICENSE = 'new BSD'
DOWNLOAD_URL = 'https://github.com/naviveztim/CDP_python'
INSTALL_REQUIRES = [
    "numpy>=1.19.5",
    "numba>=0.53.1",
    "setuptools>=58.0.4"
]
VERSION = '0.1.53'
CLASSIFIERS = ['Development Status :: 3 - Alpha',
               'Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9',
               'Programming Language :: Python :: 3.10',
               'Programming Language :: Python :: 3.11']

setup(name=NAME,
      maintainer=MAINTAINER,
      author=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
      zip_safe=False,
      classifiers=CLASSIFIERS,
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES
      )



import numpy as np
import csv
from cdp_ts.utils.logger import logger


class Dataset:

    def __init__(self, filepath: str = None
                     , delimiter: str = None
                     , class_indexes: np.array = None
                     , values: np.array = None
                     , no_indexes=False):

        if filepath is not None and delimiter is not None:
            self.class_indexes, self.values = self._load_dataset(filepath, delimiter, no_indexes)
            self.empty = True if self.class_indexes.size == 0 or self.values.size == 0 else False
        elif class_indexes is not None and values is not None:
            self.class_indexes = class_indexes
            self.values = values
            self.empty = False
        else:
            self.class_indexes = None
            self.values = None
            self.empty = True

    def iterrows(self):
        for class_index, row in zip(self.class_indexes, self.values):
            yield class_index, row

    @staticmethod
    def _load_dataset(filepath: str, delimiter: str, no_indexes: bool) -> tuple:
        """Load a UCR type data set from a local folder. """

        try:
            # Parse data file
            data = np.genfromtxt(filepath, delimiter=delimiter)

            # Extract class indexes and values from UCR like txt file
            if no_indexes:
                # There are only values-no indexes in the file
                values = data[:, 0:]
                class_indexes = np.full(values.shape[0], -1)
            else:
                # Indexes are at the beginning of every row
                values = data[:, 1:]
                class_indexes = data[:, 0]
                class_indexes = class_indexes.astype('float64').astype('int64')

        except FileNotFoundError:
            logger.info("ERROR: File not found!")
            raise
        except OSError:
            logger.info("ERROR: OS error appeared!")
            raise
        except Exception as e:
            logger.info(f"ERROR: {str(e)}")
            raise

        return class_indexes, values

    def copy(self):
        """ Makes shallow copy of Dataset object """

        new_copy = Dataset(None
                           , None
                           , np.copy(self.class_indexes)
                           , np.copy(self.values)
                           )

        return new_copy

    def apply_normalization(self):

        """Apply normalization: (x - m)/sigma on given time series """

        # Calculate the mean and standard deviation for each row
        mean_values = np.mean(self.values, axis=1, keepdims=True)
        std_values = np.std(self.values, axis=1, keepdims=True)

        # Normalize each row separately using (x - mean) / std
        self.values = (self.values - mean_values) / std_values

    def apply_compression(self, compression_factor: int):
        """ Takes averaged sample of every 'compression_factor' samples """

        new_values = []
        for x in self.values:
            rolling_mean = np.convolve(x, np.ones(compression_factor) / compression_factor, 'valid')
            new_values.append(rolling_mean[::compression_factor])
        self.values = np.array(new_values)

    def apply_derivative(self):
        """ Takes derivatives of time series in given time series"""
        self.values = np.diff(self.values, axis=1)

    def filter_by_class(self, class_index):
        """ """

        filtered_indices = np.where(self.class_indexes == class_index)[0]
        return Dataset(None, None, self.class_indexes[filtered_indices], self.values[filtered_indices])

    def to_ucr_format(self, filename: str, delimiter: str = ','):
        """ Save predicted class indexes along with given time series"""

        with open(filename, mode='w', newline='') as file:
            writer = csv.writer(file
                                , delimiter=delimiter
                                , quotechar='"'
                                , quoting=csv.QUOTE_MINIMAL)

            for class_index, row_values in zip(self.class_indexes, self.values):
                data_row = [class_index] + list(row_values)
                writer.writerow(data_row)



""" Particle Swarm Optimization (PSO) approach for finding
    shapelet that mostly separate two classes of time series"""
import sys
import numpy as np
from cdp_ts.utils.utils import assess_candidate_position
from cdp_ts.utils.dataset import Dataset
import numba
from numba import NumbaWarning
import warnings
warnings.filterwarnings("ignore", category=NumbaWarning)


class CandidateShapelet:
    """ Shapelet candidate implementation"""
    def __init__(self
                 , length: int
                 , min_velocity: float, max_velocity: float
                 , min_position: float, max_position: float):
        self.optimal_split_distance: float = 0.0
        self.best_information_gain: float = sys.float_info.min
        self.length: int = length
        self.position: np.array = np.random.uniform(min_position, max_position, length)
        self.velocity: np.array = np.random.uniform(min_velocity, max_velocity, length)
        self.best_position: np.array = self.position

    def copy(self, candidate):
        """ Copy so found the best parameters to given candidate"""

        if isinstance(candidate, CandidateShapelet):
            self.optimal_split_distance = candidate.optimal_split_distance
            self.best_information_gain = candidate.best_information_gain
            self.length = candidate.length
            self.position[:len(candidate.position)] = candidate.position
            self.velocity[:len(candidate.velocity)] = candidate.velocity
            self.best_position = candidate.best_position


class ShapeletsPso:
    """ Implementation of Particle Swarm Optimization (PSO) algorithm
        compares several candidate shapelets in a swarm. As result select the one which
        maximally separate two classes."""

    # Constants of the process (empirically found)
    W = 0.729  # inertia weight
    C1 = 1.49445  # cognitive / local weight
    C2 = 1.49445  # social / global weight

    # Stop optimization condition
    MAX_ITERATIONS = 20
    ITERATION_EPSILON = 0.0000001

    def __init__(self
                 , min_length: int, max_length: int
                 , step: int
                 , min_position: float, max_position: float
                 , min_velocity: float, max_velocity: float
                 , train_dataframe: Dataset
                 ):

        self.swarm: list = []
        self.min_position: float = min_position
        self.max_position: float = max_position
        self.min_velocity: float = min_velocity
        self.max_velocity: float = max_velocity
        self.min_particle_length: int = min_length
        self.max_particle_length: int = max_length
        self.step: int = step
        self.train_dataframe = train_dataframe

        # Init best particle
        self.best_particle: CandidateShapelet = CandidateShapelet(max_length
                                               , min_velocity, max_velocity
                                               , min_position, max_position)
        # Init swarm
        self._init_swarm()

    @numba.jit()
    def _init_swarm(self):
        """ Initialize candidate shapelets' parameters """

        for length in range(self.min_particle_length, self.max_particle_length+1, self.step):

            # Create candidate shapelet
            candidate = CandidateShapelet(length
                                          , self.min_velocity, self.max_velocity
                                          , self.min_position, self.max_position)

            # Assess candidate position
            information_gain, split_point = \
                assess_candidate_position(candidate.position
                                          , self.train_dataframe)

            # Move candidate towards best position
            if candidate.best_information_gain < information_gain:
                candidate.best_information_gain = information_gain
                candidate.best_position = candidate.position
                candidate.optimal_split_distance = split_point
                candidate.length = length

            # Create swarm of candidates
            self.swarm.append(candidate)

            # Initialize best candidate
            if candidate.best_information_gain > self.best_particle.best_information_gain:
                self.best_particle.copy(candidate)

    @numba.jit()
    def start_pso(self):
        """ Find the best shapelet that distinguishes time series from two classes"""

        new_best_gain = 0.0
        iteration = 0

        while True:

            iteration += 1
            for candidate in self.swarm:

                # Update candidate velocity
                for i in range(candidate.length):

                    r_1 = np.random.rand()
                    r_2 = np.random.rand()

                    candidate.velocity[i] = self.W * candidate.velocity[i] + \
                        self.C1*r_1*(candidate.best_position[i] - candidate.position[i]) + \
                        self.C2*r_2*(self.best_particle.position[i] - candidate.position[i])

                # Update candidate position
                candidate.position += candidate.velocity

                # Check the fitness of current candidate
                information_gain, split_point = \
                    assess_candidate_position(candidate.position
                                              , self.train_dataframe)

                # Move candidate towards best position
                if candidate.best_information_gain < information_gain:
                    candidate.best_information_gain = information_gain
                    candidate.best_position = candidate.position
                    candidate.optimal_split_distance = split_point

                # Update best particle
                if candidate.best_information_gain > self.best_particle.best_information_gain:
                    self.best_particle.copy(candidate)

            old_best_gain = new_best_gain
            new_best_gain = self.best_particle.best_information_gain

            if abs(old_best_gain - new_best_gain) <= ShapeletsPso.ITERATION_EPSILON:
                break

        self.best_particle.position = self.best_particle.position[:self.best_particle.length]